/*
 * Copyright 1993-2009 NVIDIA Corporation.  All rights reserved.
 *
 * NVIDIA Corporation and its licensors retain all intellectual property and 
 * proprietary rights in and to this software and related documentation. 
 * Any use, reproduction, disclosure, or distribution of this software 
 * and related documentation without an express license agreement from
 * NVIDIA Corporation is strictly prohibited.
 *
 * Please refer to the applicable NVIDIA end user license agreement (EULA) 
 * associated with this source code for terms and conditions that govern 
 * your use of this NVIDIA software.
 * 
 *
 *
 * This sample illustrates the usage of CUDA streams for overlapping
 * kernel execution with device/host memcopies.  The kernel is used to 
 * initialize an array to a specific value, after which the array is 
 * copied to the host (CPU) memory.  To increase performance, multiple 
 * kernel/memcopy pairs are launched asynchronously, each pair in its 
 * own stream.  Devices with Compute Capability 1.1 can overlap a kernel
 * and a memcopy as long as they are issued in different streams.  Kernels
 * are serialized.  Thus, if n pairs are launched, streamed approach
 * can reduce the memcopy cost to the (1/n)th of a single copy of the entire
 * data set.
 *
 * Additionally, this sample uses CUDA events to measure elapsed time for
 * CUDA calls.  Events are a part of CUDA API and provide a system independent
 * way to measure execution times on CUDA devices with approximately 0.5 
 * microsecond precision.
 *
 * Elapsed times are averaged over nreps repetitions (10 by default).
 *
 */

#include <stdio.h>
#include <sys/time.h>

#include <stdio.h>
#include <cutil_inline.h>
#include <cublas.h>



#define ELEMENT float
#ifdef DOUBLE
#undef ELEMENT
#define ELEMENT double
#endif


/* Host implementation of a simple version of sgemm */
static void simple_sgemm(int n, ELEMENT alpha, const ELEMENT *A, const ELEMENT *B,
		ELEMENT beta, ELEMENT *C)
{
	int i;
	int j;
	int k;
	for (i = 0; i < n; ++i) {
		for (j = 0; j < n; ++j) {
			ELEMENT prod = 0;
			for (k = 0; k < n; ++k) {
				//printf("i %d j %d k %d \n",i,j,k);
				prod += A[k + i*n] * B[j + k*n];
				//                prod += B[j + k*n];
			}
			C[j + i*n] = alpha * prod + beta * C[j + i*n];
		}
	}
}
void printMatrixFile(ELEMENT* matrix, int dim,const char * name){
	int i,j;
	FILE * fptr;

	fptr = fopen(name,"w");

	for(i=0;i<dim;++i){
		for(j=0;j<dim;++j){
			fprintf(fptr,"%f\t",matrix[i*dim + j]);
		}
		fprintf(fptr,"\n");
	}

	fclose(fptr);
}


void printMatrix(ELEMENT* matrix, int dim){
	int i,j;
	for(i=0;i<dim;++i){
		for(j=0;j<dim;++j){
			printf("%f ",matrix[i*dim + j]);
		}
		printf("\n");
	}
}

bool checkResults(ELEMENT * reference, ELEMENT * candidate, int dim){
	int i;
	float error_norm;
	float ref_norm;
	float diff;

	/* Check result against reference */
	error_norm = 0;
	ref_norm = 0;
	for (i = 0; i < dim*dim; ++i) {
		diff = reference[i] - candidate[i];
		error_norm += diff * diff;
		ref_norm += reference[i] * reference[i];
	}
	error_norm = (float)sqrt((double)error_norm);
	ref_norm = (float)sqrt((double)ref_norm);
	if (fabs(ref_norm) < 1e-7) {
		fprintf (stderr, "!!!! reference norm is 0\n");
		return EXIT_FAILURE;
	}

	printf( "Test %s\n", (error_norm / ref_norm < 1e-6f) ? "PASSED" : "FAILED");

	return  (error_norm / ref_norm < 1e-6f) ? true : false;
}

#define SIMUL(i,j) (__umul24((i),(j)))
#define ACCUMULATE(acc,i,j) (acc+=PRODUCT(i,j))
#define PRODUCT(i,j) ((i)*(j))
#define PTRGA(i,j) (((ELEMENT *)((char *)d_acol + (i)*(pitch_a)) + (j)))
#define PTRGB(i,j) (((ELEMENT *)((char *)d_brow + (i)*(pitch_b)) + (j)))
#define PTRGC(i,j) (((ELEMENT *)((char *)d_cblk + (i)*(pitch_c)) + (j)))

#ifdef DOUBLE
__global__ void mmre(ELEMENT * d_cblk, ELEMENT * d_acol, ELEMENT * d_brow, int pitch_c,int pitch_a,int pitch_b)
{
	/*shared mem segment used to cache A and B*/
	__shared__ ELEMENT shb[GAMMA*(MU+1)];
	__shared__ ELEMENT sha[(GAMMA+1)*XI];

	volatile const unsigned int tx = (threadIdx.x & (31));
	const unsigned int atx = (threadIdx.x & (15));
	volatile const unsigned int ty = (threadIdx.x >> 5);
	const unsigned int aty = (threadIdx.x >> 4);
	const unsigned int bx = (blockIdx.x)*ALPHAX;
	const unsigned int by = (blockIdx.y)*ALPHAY;

	/*pointer to the sub array that needs to be updated*/
	ELEMENT * my_gc = PTRGC(by + ty,bx + tx);
	ELEMENT * my_ga = PTRGA(by + aty,atx);
	ELEMENT * my_gb = PTRGB(ty,bx + tx);
	unsigned int ii, jj,kk,k;
	for(jj = 0; jj< ALPHAX; jj+=MU){
		ELEMENT * my_curblock = my_gc;
		ELEMENT * my_lga = my_ga;
		for(ii = 0; ii < ALPHAY; ii+=XI){
			/*registers used to store accumulations from A and B*/
			/*we load C in registers*/

			ELEMENT * my_cline = my_curblock;
			ELEMENT acc_00=my_cline[0];
			ELEMENT acc_01=my_cline[32];
			my_cline += GAMMA*ALPHA;
			ELEMENT acc_10=my_cline[0];
			ELEMENT acc_11=my_cline[32];
			ELEMENT * my_kga = my_lga;
			ELEMENT * my_kgb = my_gb;
			volatile ELEMENT* my_shb = shb + SIMUL(ty,(MU+1)) + tx;
			for(kk=0;kk<BETA;kk+=GAMMA){
				__syncthreads();
				ELEMENT* my_sha = sha + SIMUL(aty,(GAMMA+1)) + atx;

				/*load A and B in shared mem */
				my_sha[0] = my_kga[0];

				my_shb[0] = my_kgb[0];
				my_shb[32] = my_kgb[32];

				my_shb = shb + tx;
				my_sha = ty*(GAMMA+1) + sha;

				my_kga += GAMMA;
				my_kgb += GAMMA*ALPHA;
				__syncthreads();
				/*accumulate A and B in acc_i*/

				for(k=0;k<GAMMA;k++){
					ELEMENT a_k = my_sha[k];
					ACCUMULATE(acc_00,a_k,my_shb[0]);
					ACCUMULATE(acc_01,a_k,my_shb[32]);

					a_k = my_sha[k+GAMMA*(GAMMA+1)];
					ACCUMULATE(acc_10,a_k,my_shb[0]);
					ACCUMULATE(acc_11,a_k,my_shb[32]);
					my_shb += (MU+1);
				}


			}

			my_curblock[0] = acc_00;
			my_curblock[32] = acc_01;
			my_curblock += GAMMA*ALPHA;
			my_curblock[0] = acc_10;
			my_curblock[32] = acc_11;
			my_curblock += GAMMA*ALPHA;
			my_lga += XI * BETA;
		}
		my_gb += MU;
		my_gc += MU;	
	}
}
#else

__global__ void mmre(ELEMENT * d_cblk, ELEMENT * d_acol, ELEMENT * d_brow, int pitch_c,int pitch_a,int pitch_b)
{
	/*shared mem segment used to cache A and B*/
	__shared__ ELEMENT shb[GAMMA*(MU+1)];
	__shared__ ELEMENT sha[(GAMMA+1)*XI];

	volatile const unsigned int tx = (threadIdx.x & (31));
	const unsigned int atx = (threadIdx.x & (15));
	volatile const unsigned int ty = (threadIdx.x >> 5);
	const unsigned int aty = (threadIdx.x >> 4);
	const unsigned int bx = (blockIdx.x)*ALPHAX;
	const unsigned int by = (blockIdx.y)*ALPHAY;

	/*pointer to the sub array that needs to be updated*/
	ELEMENT * my_gc = PTRGC(by + ty,bx + tx);
	ELEMENT * my_ga = PTRGA(by + aty,atx);
	ELEMENT * my_gb = PTRGB(ty,bx + tx);
	unsigned int ii, jj,kk,k;
	for(jj = 0; jj< ALPHAX; jj+=MU){
		ELEMENT * my_curblock = my_gc;
		ELEMENT * my_lga = my_ga;
		for(ii = 0; ii < ALPHAY; ii+=XI){
			/*registers used to store accumulations from A and B*/
			/*we load C in registers*/

			ELEMENT * my_cline = my_curblock;
			ELEMENT acc_00=my_cline[0];
			ELEMENT acc_01=my_cline[32];
			ELEMENT acc_02=my_cline[64];
			my_cline += GAMMA*ALPHA;
			ELEMENT acc_10=my_cline[0];
			ELEMENT acc_11=my_cline[32];
			ELEMENT acc_12=my_cline[64];
			my_cline += GAMMA*ALPHA;
			ELEMENT acc_20=my_cline[0];
			ELEMENT acc_21=my_cline[32];
			ELEMENT acc_22=my_cline[64];
			my_cline += GAMMA*ALPHA;
			ELEMENT acc_30=my_cline[0];
			ELEMENT acc_31=my_cline[32];
			ELEMENT acc_32=my_cline[64];
			my_cline += GAMMA*ALPHA;
			ELEMENT acc_40=my_cline[0];
			ELEMENT acc_41=my_cline[32];
			ELEMENT acc_42=my_cline[64];
			ELEMENT * my_kga = my_lga;
			ELEMENT * my_kgb = my_gb;
			volatile ELEMENT* my_shb = shb + SIMUL(ty,(MU+1)) + tx;
			for(kk=0;kk<BETA;kk+=GAMMA){
				__syncthreads();
				ELEMENT* my_sha = sha + SIMUL(aty,(GAMMA+1)) + atx;

				/*load A and B in shared mem */
				my_sha[0] = my_kga[0];
				my_sha[32 * (GAMMA+1)] = my_kga[32*BETA];
				if(aty<16){
					my_sha[64 * (GAMMA+1)] = my_kga[64*BETA];
				}

				my_shb[0] = my_kgb[0];
				my_shb[32] = my_kgb[32];
				my_shb[64] = my_kgb[64];

				my_shb = shb + tx;
				my_sha = ty*(GAMMA+1) + sha;

				my_kga += GAMMA;
				my_kgb += GAMMA*ALPHA;
				__syncthreads();
				/*accumulate A and B in acc_i*/

				for(k=0;k<GAMMA;k++){
					ELEMENT a_k = my_sha[k];

					ACCUMULATE(acc_00,a_k,my_shb[0]);
					ACCUMULATE(acc_01,a_k,my_shb[32]);
					ACCUMULATE(acc_02,a_k,my_shb[64]);

					a_k = my_sha[k+GAMMA*(GAMMA+1)];
					ACCUMULATE(acc_10,a_k,my_shb[0]);
					ACCUMULATE(acc_11,a_k,my_shb[32]);
					ACCUMULATE(acc_12,a_k,my_shb[64]);

					a_k = my_sha[k+2*GAMMA*(GAMMA+1)];
					ACCUMULATE(acc_20,a_k,my_shb[0]);
					ACCUMULATE(acc_21,a_k,my_shb[32]);
					ACCUMULATE(acc_22,a_k,my_shb[64]);

					a_k = my_sha[k+3*GAMMA*(GAMMA+1)];
					ACCUMULATE(acc_30,a_k,my_shb[0]);
					ACCUMULATE(acc_31,a_k,my_shb[32]);
					ACCUMULATE(acc_32,a_k,my_shb[64]);

					a_k = my_sha[k+4*GAMMA*(GAMMA+1)];
					ACCUMULATE(acc_40,a_k,my_shb[0]);
					ACCUMULATE(acc_41,a_k,my_shb[32]);
					ACCUMULATE(acc_42,a_k,my_shb[64]);

					my_shb += (MU+1);
				}


			}

			my_curblock[0] = acc_00;
			my_curblock[32] = acc_01;
			my_curblock[64] = acc_02;
			my_curblock += GAMMA*ALPHA;
			my_curblock[0] = acc_10;
			my_curblock[32] = acc_11;
			my_curblock[64] = acc_12;
			my_curblock += GAMMA*ALPHA;
			my_curblock[0] = acc_20;
			my_curblock[32] = acc_21;
			my_curblock[64] = acc_22;
			my_curblock += GAMMA*ALPHA;
			my_curblock[0] = acc_30;
			my_curblock[32] = acc_31;
			my_curblock[64] = acc_32;
			my_curblock += GAMMA*ALPHA;
			my_curblock[0] = acc_40;
			my_curblock[32] = acc_41;
			my_curblock[64] = acc_42;
			my_curblock += GAMMA*ALPHA;

			my_lga += XI * BETA;
		}
		my_gb += MU;
		my_gc += MU;	
	}
}
#endif

int main(int argc, char *argv[])
{

    if( cutCheckCmdLineFlag(argc, (const char**)argv, "device") )
        cutilDeviceInit(argc, argv);
    else
        cudaSetDevice( cutGetMaxGflopsDeviceId() );

	struct timeval start, finish,start2,finish2 ;
	int msec_cpu,msec_gpu,msec_cublas;

	int mu = MU;
	int alpha = ALPHA;
	int beta = BETA;

	int cuda_device = 0;
	int nstreams = 2;               // number of streams for CUDA calls
	dim3 threads, blocks;           // kernel launch configuration
	float elapsed_time, time_memcpy, time_kernel;   // timing variables

	// allocate host memory
	ELEMENT * h_c = 0;
	ELEMENT * h_ref_c = 0;
	ELEMENT * h_a = 0;
	ELEMENT * h_b = 0;
	ELEMENT * h_cublas = 0; 
#ifdef CPU
	h_ref_c = (ELEMENT*)malloc(PADN*PADN*sizeof(ELEMENT));
        if(h_ref_c ==0)
		return EXIT_FAILURE;
#endif

	    cutilSafeCall(cudaMallocHost((void**)&h_a, PADN*PADN*sizeof(ELEMENT))) ; 
	    cutilSafeCall(cudaMallocHost((void**)&h_b, PADN*PADN*sizeof(ELEMENT))) ; 

#if defined(MMRE) or defined(TOLEDO)
	    cutilSafeCall(cudaMallocHost((void**)&h_c, PADN*PADN*sizeof(ELEMENT))) ; 
#endif

#ifdef CUBLAS
	  cutilSafeCall(cudaMallocHost((void**)&h_cublas, PADN*PADN*sizeof(ELEMENT))) ; 
#endif

	srand((unsigned int)time((time_t *)NULL));

	/* Fill the matrices with test data */
	for (int i = 0; i < PADN; i++) {
		for (int j = 0; j < PADN; j++) {
			if(i<ORDER && j<ORDER){
				h_a[i*PADN +j] = rand() / (float)RAND_MAX;
				h_b[i*PADN +j] = rand() / (float)RAND_MAX;
				ELEMENT c = rand() / (float)RAND_MAX;
#if defined(MMRE) or defined(TOLEDO)
				h_c[i*PADN + j] = c;
#endif

#ifdef CPU
				h_ref_c[i*PADN +j] = c;
#endif

#ifdef CUBLAS
				h_cublas[i*PADN +j] = c;
#endif
			}
			else{
				h_a[i*PADN +j] = 0.0;
				h_b[i*PADN +j] = 0.0;
#if defined(MMRE) or defined(TOLEDO)
				h_c[i*PADN + j] = 0.0;
#endif
#ifdef CPU
				h_ref_c[i*PADN +j] = 0.0;
#endif

#ifdef CUBLAS
				h_cublas[i*PADN + j] = 0.0;
#endif
			}
		}
	}
#ifdef PRINT 
	printMatrixFile(h_a,PADN,"CIN.txt");
#endif


#ifdef CPU 
	printf("running sgemm on CPU\n");
	gettimeofday (&start, NULL);
	simple_sgemm(PADN, 1.0f, h_a, h_b,1.0f,h_ref_c);
	gettimeofday (&finish, NULL);
	msec_cpu = finish.tv_sec * 1000 + finish.tv_usec / 1000;
	msec_cpu -= start.tv_sec * 1000 + start.tv_usec / 1000;
#endif



#ifdef PRINT
#ifdef CPU
	printMatrixFile(h_ref_c,PADN,"COUT_REF.txt");
#endif
#endif

#ifdef CPU
	free(h_ref_c);
#endif


	ELEMENT * d_cblk = 0;
	ELEMENT ** d_acol = 0;
	ELEMENT ** d_brow = 0;
	ELEMENT ** h_acol = 0;
	ELEMENT ** h_brow = 0;

	int maj_step = alpha;
	int min_step = beta;

	gettimeofday (&start, NULL);

	// allocate device memory
	d_acol = (ELEMENT **)malloc(nstreams * sizeof(ELEMENT *));
	d_brow = (ELEMENT **)malloc(nstreams * sizeof(ELEMENT *));
#ifdef BESTOVERLAP    
	h_acol = (ELEMENT **)malloc(nstreams * sizeof(ELEMENT *));
	h_brow = (ELEMENT **)malloc(nstreams * sizeof(ELEMENT *));
#endif

	int * pitch_a;
	int * pitch_b;
	int pitch_c;
	pitch_a = (int *)malloc(nstreams * sizeof(int));
	pitch_b = (int *)malloc(nstreams * sizeof(int));


#ifdef MMRE
	cutilSafeCall(cudaMallocPitch((void**)&d_cblk,(size_t*)&pitch_c, (size_t)alpha*sizeof(ELEMENT),(size_t)alpha) );
	for(int i = 0; i < nstreams; i++){

#ifdef BESTOVERLAP        
		cutilSafeCall( cudaMallocHost((void**)&h_acol[i], alpha*beta*sizeof(ELEMENT)) ); 
		cutilSafeCall( cudaMallocHost((void**)&h_brow[i], alpha*beta*sizeof(ELEMENT)) ); 
#endif

		cutilSafeCall( cudaMallocPitch((void**)&d_acol[i],(size_t*)&pitch_a[i],(size_t)beta*sizeof(ELEMENT),(size_t)alpha) );
		cutilSafeCall( cudaMallocPitch((void**)&d_brow[i],(size_t*)&pitch_b[i],(size_t)alpha*sizeof(ELEMENT),(size_t)beta) );
		printf("pitch B %d, pitch A %d, pitch C %d\n",pitch_b[i],pitch_a[i],pitch_c);
	}

	// allocate and initialize an array of stream handles
	cudaStream_t *streams = (cudaStream_t*) malloc(nstreams * sizeof(cudaStream_t));
	for(int i = 0; i < nstreams; i++)
		cutilSafeCall( cudaStreamCreate(&(streams[i])) );

	// create CUDA event handles
	cudaEvent_t start_event, stop_event;
	cutilSafeCall( cudaEventCreate(&start_event) );
	cutilSafeCall( cudaEventCreate(&stop_event) );

	//////////////////////////////////////////////////////////////////////
	// time execution with nstreams streams
	threads=dim3(512,1);
	blocks=dim3(5,6);
	cutilSafeCall(cudaEventRecord(start_event, 0));



	for(int REPI = 0 ; REPI < REPEAT; REPI++)
	{

		for(int stepi = 0; stepi<PADN; stepi+=maj_step){

			for(int REPJ = 0 ; REPJ < REPEAT; REPJ++)
			{

				for(int stepj= 0; stepj < PADN; stepj+=maj_step){
					/* load square block of C */
					cutilSafeCall(	cudaMemcpy2D(d_cblk,pitch_c ,&h_c[stepi * PADN + stepj],PADN*sizeof(ELEMENT) , alpha*sizeof(ELEMENT), alpha , cudaMemcpyHostToDevice));

					for(int REPK = 0 ; REPK < REPEAT; REPK++)
					{
						int stepk =0;
#ifdef BESTOVERLAP
						cutilSafeCall(	cudaMemcpy2D(h_acol[stepk%nstreams],beta*sizeof(ELEMENT) ,&h_a[stepi*PADN + (stepk)*min_step],PADN*sizeof(ELEMENT), beta*sizeof(ELEMENT), alpha , cudaMemcpyHostToHost));
						for(int lk = 0; lk < beta; lk++){
							memcpy(h_brow[stepk&1] + lk*alpha,&h_b[((stepk)*min_step+lk)*PADN + stepj], alpha*sizeof(ELEMENT));
						}

						cutilSafeCall(	cudaMemcpyAsync(d_acol[stepk%nstreams], h_acol[stepk%nstreams],alpha* beta*sizeof(ELEMENT) , cudaMemcpyHostToDevice,streams[stepk%nstreams]));
						cutilSafeCall(	cudaMemcpyAsync(d_brow[stepk%nstreams], h_brow[stepk%nstreams],alpha* beta*sizeof(ELEMENT) , cudaMemcpyHostToDevice,streams[stepk%nstreams]));

						stepk++;

						while(stepk< PADN/min_step){
							mmre<<<blocks, threads, 0, streams[(stepk-1)%nstreams]>>>(d_cblk,  d_acol[(stepk-1)%nstreams] ,  d_brow[(stepk-1)%nstreams], pitch_c, pitch_a[(stepk-1)%nstreams],pitch_b[(stepk-1)%nstreams]);
							cutilSafeCall(	cudaMemcpy2D(h_acol[stepk%nstreams],beta*sizeof(ELEMENT) ,&h_a[stepi*PADN + (stepk)*min_step],PADN*sizeof(ELEMENT), beta*sizeof(ELEMENT), alpha , cudaMemcpyHostToHost));
							for(int lk = 0; lk < beta; lk++){
								memcpy(h_brow[stepk&1] + lk*alpha,&h_b[((stepk)*min_step+lk)*PADN + stepj], alpha*sizeof(ELEMENT));
							}

							cutilSafeCall(	cudaMemcpyAsync(d_acol[stepk%nstreams], h_acol[stepk%nstreams],alpha* beta*sizeof(ELEMENT) , cudaMemcpyHostToDevice,streams[stepk%nstreams]));
							cutilSafeCall(	cudaMemcpyAsync(d_brow[stepk%nstreams], h_brow[stepk%nstreams],alpha* beta*sizeof(ELEMENT) , cudaMemcpyHostToDevice,streams[stepk%nstreams]));

                            stepk++;
						}

						mmre<<<blocks, threads, 0, streams[(stepk-1)%nstreams]>>>(d_cblk,  d_acol[(stepk-1)&1] ,  d_brow[(stepk-1)&1], pitch_c, pitch_a[(stepk-1)&1],pitch_b[(stepk-1)&1]);

#else

						cudaMemcpy2DAsync(d_brow[stepk&1],pitch_b[stepk&1], &h_b[(stepk)*min_step*PADN + stepj],PADN*sizeof(ELEMENT) , alpha*sizeof(ELEMENT),beta  , cudaMemcpyHostToDevice,streams[stepk%nstreams]);
						cudaMemcpy2DAsync(d_acol[stepk&1],pitch_a[stepk&1], &h_a[stepi*PADN + (stepk)*min_step],PADN*sizeof(ELEMENT) , beta*sizeof(ELEMENT),alpha  , cudaMemcpyHostToDevice,streams[stepk%nstreams]);

						stepk++;
						
                        while(stepk< PADN/min_step){
						    mmre<<<blocks, threads, 0, streams[(stepk-1)%nstreams]>>>(d_cblk,  d_acol[(stepk-1)&1] ,  d_brow[(stepk-1)&1], pitch_c, pitch_a[(stepk-1)&1],pitch_b[(stepk-1)&1]);
                            cudaMemcpy2DAsync(d_brow[stepk&1],pitch_b[stepk&1], &h_b[(stepk)*min_step*PADN + stepj],PADN*sizeof(ELEMENT) , alpha*sizeof(ELEMENT),beta  , cudaMemcpyHostToDevice,streams[stepk%nstreams]);
					        cudaMemcpy2DAsync(d_acol[stepk&1],pitch_a[stepk&1], &h_a[stepi*PADN + (stepk)*min_step],PADN*sizeof(ELEMENT) , beta*sizeof(ELEMENT),alpha  , cudaMemcpyHostToDevice,streams[stepk%nstreams]);

						    stepk++;
                        }


						mmre<<<blocks, threads, 0, streams[(stepk-1)%nstreams]>>>(d_cblk,  d_acol[(stepk-1)&1] ,  d_brow[(stepk-1)&1], pitch_c, pitch_a[(stepk-1)&1],pitch_b[(stepk-1)&1]);
#endif
					} /*end REPK*/
                	cutilSafeCall(cudaStreamSynchronize(streams[0]));
        			cutilSafeCall(cudaStreamSynchronize(streams[1]));


					/* get back square block of C */
					cutilSafeCall(	cudaMemcpy2D(&h_c[stepi * PADN + stepj],PADN*sizeof(ELEMENT) , d_cblk,pitch_c ,alpha*sizeof(ELEMENT),alpha  , cudaMemcpyDeviceToHost));

				}
			} /*end REPJ*/

		}
	} /*end REPI*/
	cudaThreadSynchronize();
	cudaEventRecord(stop_event, 0);
	cudaEventSynchronize(stop_event);
	cutilSafeCall(cudaEventElapsedTime(&elapsed_time, start_event, stop_event));
	printf("elapsed time \t%.2f\n", elapsed_time);

	// release resources
	for(int i = 0; i < nstreams; i++){
		cudaStreamDestroy(streams[i]);
		cudaFree(d_acol[i]);
		cudaFree(d_brow[i]);
#ifdef BESTOVERLAP
		cudaFreeHost(h_acol[i]);
		cudaFreeHost(h_brow[i]);
#endif
	}
	free(h_acol);
	free(h_brow);

	cudaFree(d_cblk);

	cudaEventDestroy(start_event);
	cudaEventDestroy(stop_event);

	gettimeofday (&finish, NULL);

	msec_gpu = finish.tv_sec * 1000 + finish.tv_usec / 1000;
	msec_gpu -= start.tv_sec * 1000 + start.tv_usec / 1000;
#endif



#ifdef TOLEDO

	cutilSafeCall( cudaMallocPitch((void**)&d_cblk,(size_t*)&pitch_c, (size_t)alpha*sizeof(ELEMENT),(size_t)alpha)) ;
	cutilSafeCall( cudaMallocPitch((void**)&d_acol[0],(size_t*)&pitch_a[0],(size_t)beta*sizeof(ELEMENT),(size_t)alpha) );
	cutilSafeCall( cudaMallocPitch((void**)&d_brow[0],(size_t*)&pitch_b[0],(size_t)alpha*sizeof(ELEMENT),(size_t)beta) );

	// create CUDA event handles
	cudaEvent_t start_event, stop_event;
	cutilSafeCall( cudaEventCreate(&start_event) );
	cutilSafeCall( cudaEventCreate(&stop_event) );

	//////////////////////////////////////////////////////////////////////
	// time execution with nstreams streams

	threads=dim3(512,1);

	blocks=dim3(5,6);
	cudaEventRecord(start_event, 0);

	for(int REPI = 0 ; REPI < REPEAT; REPI++)
	{
		for(int stepi = 0; stepi<PADN; stepi+=maj_step){
			for(int REPJ = 0 ; REPJ < REPEAT; REPJ++)
			{
				for(int stepj= 0; stepj < PADN; stepj+=maj_step){
					/* load square block of C */
					cudaMemcpy2D(d_cblk,pitch_c ,&h_c[stepi * PADN + stepj],PADN*sizeof(ELEMENT) , alpha*sizeof(ELEMENT), alpha , cudaMemcpyHostToDevice);
					for(int REPK = 0 ; REPK < REPEAT; REPK++){
						int stepk =0;
						while(stepk< PADN){
							cudaMemcpy2D(d_brow[0],pitch_b[0], &h_b[stepk*PADN + stepj],PADN*sizeof(ELEMENT) , alpha*sizeof(ELEMENT),alpha  , cudaMemcpyHostToDevice);
							cudaMemcpy2D(d_acol[0],pitch_a[0], &h_a[stepi*PADN + stepk],PADN*sizeof(ELEMENT) , alpha*sizeof(ELEMENT),alpha  , cudaMemcpyHostToDevice);
							mmre<<<blocks, threads, 0>>>(d_cblk,  d_acol[0] ,  d_brow[0], pitch_c, pitch_a[0],pitch_b[0]);
							stepk+=min_step;
						}
					} /*end REPK*/
					/* get back square block of C */
					cudaMemcpy2D(&h_c[stepi * PADN + stepj],PADN*sizeof(ELEMENT) , d_cblk,pitch_c ,alpha*sizeof(ELEMENT),alpha  , cudaMemcpyDeviceToHost);
				}
			} /*end REPJ*/
		}
	} /*end REPI*/
	cudaThreadSynchronize();
	cudaEventRecord(stop_event, 0);
	cudaEventSynchronize(stop_event);
	cutilSafeCall(cudaEventElapsedTime(&elapsed_time, start_event, stop_event));
	printf("elapsed time \t%.2f\n", elapsed_time);

	// release resources
	cudaFree(d_acol[0]);
	cudaFree(d_brow[0]);
	cudaFree(d_cblk);

	cudaEventDestroy(start_event);
	cudaEventDestroy(stop_event);


	gettimeofday (&finish, NULL);

	msec_gpu = finish.tv_sec * 1000 + finish.tv_usec / 1000;
	msec_gpu -= start.tv_sec * 1000 + start.tv_usec / 1000;
#endif

	free(pitch_a);
	free(pitch_b);

#ifdef CUBLAS
	cublasStatus status;

	/* Initialize CUBLAS */
	printf("simpleCUBLAS test running..\n");

	gettimeofday (&start, NULL);
	status = cublasInit();
	if (status != CUBLAS_STATUS_SUCCESS) {
		fprintf (stderr, "!!!! CUBLAS initialization error\n");
		return EXIT_FAILURE;
	}

	/* Allocate device memory for the matrices */
	status = cublasAlloc(CUBLAS_STEP*CUBLAS_STEP, sizeof(ELEMENT), (void**)&d_acol[0]);
	if (status != CUBLAS_STATUS_SUCCESS) {
		fprintf (stderr, "!!!! device memory allocation error (A)\n");
		return EXIT_FAILURE;
	}
	status = cublasAlloc(CUBLAS_STEP*CUBLAS_STEP, sizeof(ELEMENT), (void**)&d_brow[0]);
	if (status != CUBLAS_STATUS_SUCCESS) {
		fprintf (stderr, "!!!! device memory allocation error (B)\n");
		return EXIT_FAILURE;
	}
	status = cublasAlloc(CUBLAS_STEP*CUBLAS_STEP, sizeof(ELEMENT), (void**)&d_cblk);
	if (status != CUBLAS_STATUS_SUCCESS) {
		fprintf (stderr, "!!!! device memory allocation error (C)\n");
		return EXIT_FAILURE;
	}

	/* Clear last error */
	cublasGetError();

	for(int REPI = 0 ; REPI < REPEAT; REPI++)
	{
		for(int i=0;i<PADN;i+=CUBLAS_STEP){
			for(int REPJ = 0 ; REPJ < REPEAT; REPJ++)
			{
				for(int j=0;j<PADN;j+=CUBLAS_STEP){
					status = cublasSetMatrix (CUBLAS_STEP,CUBLAS_STEP, sizeof(ELEMENT),&h_cublas[i*PADN + j ],PADN,d_cblk,CUBLAS_STEP);
					if (status != CUBLAS_STATUS_SUCCESS) {
						fprintf (stderr, "!!!! device access error (write C) i %d j %d\n",i,j);
						return EXIT_FAILURE;
					}
					for(int REPK = 0 ; REPK < REPEAT; REPK++){
						for(int k=0;k<PADN;k+=CUBLAS_STEP){
							status = cublasSetMatrix (CUBLAS_STEP,CUBLAS_STEP, sizeof(ELEMENT),&h_a[k + i*PADN],PADN,d_acol[0],CUBLAS_STEP);
							if (status != CUBLAS_STATUS_SUCCESS) {
								fprintf (stderr, "!!!! device access error (write A) i %d j %d k %d\n",i,j,k);
								return EXIT_FAILURE;
							}
							status = cublasSetMatrix (CUBLAS_STEP,CUBLAS_STEP, sizeof(ELEMENT),&h_b[k*PADN + j],PADN,d_brow[0],CUBLAS_STEP);
							if (status != CUBLAS_STATUS_SUCCESS) {
								fprintf (stderr, "!!!! device access error (write B) i %d j %d k %d\n",i,j,k);
								return EXIT_FAILURE;
							}
							/* Performs operation using cublas */
#ifndef DOUBLE
							cublasSgemm('n', 'n', CUBLAS_STEP, CUBLAS_STEP, CUBLAS_STEP, 1, d_brow[0], CUBLAS_STEP, d_acol[0], CUBLAS_STEP, 1, d_cblk, CUBLAS_STEP);
#else
							cublasDgemm('n', 'n', CUBLAS_STEP, CUBLAS_STEP, CUBLAS_STEP, 1, d_brow[0], CUBLAS_STEP, d_acol[0], CUBLAS_STEP, 1, d_cblk, CUBLAS_STEP);
#endif
							status = cublasGetError();
							if (status != CUBLAS_STATUS_SUCCESS) {
								fprintf (stderr, "!!!! kernel execution error.\n");
								return EXIT_FAILURE;
							}
						}
					}/*end REPK*/
					/* Read the result back */
					status = cublasGetMatrix (CUBLAS_STEP,CUBLAS_STEP, sizeof(ELEMENT),d_cblk,CUBLAS_STEP,&h_cublas[i*PADN + j ],PADN);
					if (status != CUBLAS_STATUS_SUCCESS) {
						fprintf (stderr, "!!!! device access error (read C)\n");
						return EXIT_FAILURE;
					}
				}
			}/*end REPJ*/
		}
	}/*end REPI*/
	cudaThreadSynchronize();

	/* Memory clean up */
	status = cublasFree(d_acol[0]);
	if (status != CUBLAS_STATUS_SUCCESS) {
		fprintf (stderr, "!!!! memory free error (A)\n");
		return EXIT_FAILURE;
	}
	status = cublasFree(d_brow[0]);
	if (status != CUBLAS_STATUS_SUCCESS) {
		fprintf (stderr, "!!!! memory free error (B)\n");
		return EXIT_FAILURE;
	}
	status = cublasFree(d_cblk);
	if (status != CUBLAS_STATUS_SUCCESS) {
		fprintf (stderr, "!!!! memory free error (C)\n");
		return EXIT_FAILURE;
	}

	/* Shutdown */
	status = cublasShutdown();
	if (status != CUBLAS_STATUS_SUCCESS) {
		fprintf (stderr, "!!!! shutdown error (A)\n");
		return EXIT_FAILURE;
	}
	gettimeofday (&finish, NULL);
	msec_cublas = finish.tv_sec * 1000 + finish.tv_usec / 1000;
	msec_cublas -= start.tv_sec * 1000 + start.tv_usec / 1000;
#endif

	free(d_acol);
	free(d_brow);

#ifdef PRINT
#ifdef MMRE
	printMatrixFile(h_c,PADN,"COUT_MMRE.txt");
#endif

#ifdef TOLEDO
	printMatrixFile(h_c,PADN,"COUT_TOLEDO.txt");
#endif

#ifdef CUBLAS
	printMatrixFile(h_cublas,PADN,"COUT_CUBLAS.txt");
#endif
#endif

	printf("Freeing resources\n");
	cudaFreeHost(h_a);
	cudaFreeHost(h_b);
#if defined(MMRE) or defined(TOLEDO)
	cudaFreeHost(h_c);
#endif
#ifdef CUBLAS
	cudaFreeHost(h_cublas);
#endif


#ifdef CPU
	printf("CPU time %d ms\n",msec_cpu);
#endif
#if defined(MMRE) or defined(TOLEDO)
    printf("GPU time %d ms\n",msec_gpu);
#endif
#ifdef CUBLAS
    printf("CUBLAS time %d ms\n",msec_cublas);
#endif
   	cudaThreadExit();
}


